(** Canvas.

A canvas widget is composed of a root canvas {!class-group}. A
group can contain {!class-full_item}s (a {!class-group}s is
a [fullitem] too).

Coordinates of an item are relative to its parent group.

All item classes takes optional [x] and [y] arguments to define
their coordinates relatively to their group, with default value [0].
The group can be given with the optional [group] argument,
or set with the [set_group] method. These three arguments are
also handled by {{!convenient_functions}convenient functions
to create items}.
*)

open Tsdl

(**/**)
module Bounds = Set.Make
  (struct
     type t = int * Oid.t
     let compare (n1, id1) (n2, id2) =
       match Stdlib.compare n1 n2 with
       | 0 -> Oid.compare id1 id2
       | n -> n
   end)
(**/**)

(** {2 Items} *)

(** This virtual class defines the item interface. Most of the
  virtual methods correspond to methods of {!Widget.class-widget}
  and are defined by inherting from {!Widget.class-widget} or other
  widgets. *)
class virtual item ?(group:group option) ?(x=0) ?(y=0) () =
  let init_group = group in
  object(self)
    method as_item = (self :> item)
    method virtual id : Oid.t

    (**/**)
    method virtual set_geometry: G.t -> unit
    val virtual mutable g : G.t
    (**/**)
    method virtual geometry : G.t

    (**/**)
    val virtual mutable frozen : bool
    (**/**)

    method virtual get_p : 'a. 'a Props.prop -> 'a
    method virtual coerce : Widget.widget
    method virtual wtree : Widget.widget Widget.tree
    method virtual me : string

    (**/**)
    method virtual set_parent :
      ?with_rend:((Sdl.renderer -> unit Lwt.t) -> unit Lwt.t)
      -> Widget.widget option -> unit
    val virtual mutable min_width : int option
    val virtual mutable min_height : int option
    method virtual private min_width_ : int
    method virtual private min_height_ : int
    (**/**)

    method virtual min_width : int
    method virtual min_height : int
    method virtual as_full_item : full_item

    (**/**)
    method update_bounds ~(recur:bool) = ()
    (**/**)

    (** {2 Coordinates} *)

    method x = self#geometry.x
    method y = self#geometry.y
    method set_x (x_:int) = self#set_xy (x_, self#y)
    method set_y (y_:int) = self#set_xy (self#x, y_)
    method set_xy (x,y) =
      Log.debug (fun m -> m "%s#set_xy (%d,%d)" self#me x y);
      let g0 = self#geometry in
      self#set_geometry { g0 with x ; y } ;
      let g1 = self#geometry in
      match group with
      | None -> ()
      | Some group ->
          group#set_item_hbound g0 self#id g1 ;
          group#set_item_vbound g0 self#id g1

    method move ?(x=self#x) ?(y=self#y) () = self#set_xy (x,y)

    method to_canvas_coords ~x ~y =
      match group with
      | None -> (x, y)
      | Some g -> g#to_canvas_coords ~x ~y

    (** {2 Group} *)

    (**/**)
    val mutable group = (None : group option)
    (**/**)
    method group = group
    method set_group g =
      Option.iter (fun (g:group) -> g#remove_item (self#as_full_item)) group ;
      group <- g

    (**/**)
    (** this method should be used only is the group class, to prevent
       a recursive call to remove_item, trying to remove the item
       twice from the same group, issuing a warning. *)
    method set_group_none = group <- None
    (**/**)

    (** {2 Accessing items} *)

    method on_items_at (f:full_item->bool Lwt.t) ~(x:int) ~(y:int) = Lwt.return_false
    method get_item_at ~x ~y =
      if G.inside ~x ~y self#geometry then Some self#as_full_item else None

    method get_leaf_items_at acc ~x ~y =
      match self#get_item_at ~x ~y with
      | None -> acc
      | Some i -> i::acc

    (**/**)

    method virtual render : layer:Layer.t -> Sdl.renderer ->
      offset:(int*int) -> G.t -> unit

    method need_render ~layer geom =
      if not frozen then
        match group with
        | None -> ()
        | Some group ->
            Log.debug (fun m -> m "item%s#need_render: geom=%a"
               self#me G.pp geom);
            group#child_need_render ~layer geom

    method need_resize =
      let g0 = g in
      min_width <- None ;
      min_height <- None ;
      let w = self#min_width in
      let h = self#min_height in
      self#set_geometry { g0 with w ; h };
      let g1 = self#geometry in
      Log.debug (fun m -> m "%s[item]#need_resize: g0=%a g1=%a"
        self#me G.pp g0 G.pp g1);
        match group with
        | None -> ()
        | Some group ->
            group#set_item_hbound g0 self#id g1 ;
            group#set_item_vbound g0 self#id g1

    initializer
      match init_group with
      | None -> ()
      | Some group -> group#add_item ?x:(Some x) ?y:(Some y) (self#as_full_item)

  end

(** A [full_item] is a {!Widget.class-widget} with item interface.*)
and full_item ?class_ ?name ?props ?group ?x ?y () =
  object(self)
    inherit Widget.widget ?class_ ?name ?props () as widget
    inherit item ?group ?x ?y () as super
    method as_full_item = (self :> full_item)
    (**/**)
    method! need_resize =
      min_width <- None ;
      min_height <- None ;
      match group with
      | None -> widget#need_resize
      | _ -> (* child should call set_item_vbound and set_item_hbound *)
          (*Log.warn
           (fun m -> m "%s#child_need_resize should not be called when item has a group"
             self#me)*)
           ()
  end

(** a [container_item] is a {!Container.class-container} with item interface. *)
and ['a] container_item ?class_ ?name ?props ?group ?x ?y () =
  object(self)
    inherit ['a] Container.container ?class_ ?name ?props () as widget
    inherit item ?group ?x ?y () as super
    method as_full_item = ((self :> 'a container_item) :> full_item)
    (**/**)
    method need_resize =
      min_width <- None ;
      min_height <- None ;
      match group with
      | None -> widget#need_resize
      | _ -> (* child should call set_item_vbound and set_item_hbound *)
          (*Log.warn
           (fun m -> m "%s#child_need_resize should not be called when item has a group"
             self#me)*)
          ()
  end

(** A group is a \[[full_item]\] {!class-container_item}.
  Its default [class_] is ["canvas_group"]. *)
and group ?(class_="canvas_group") ?name ?props ?group ?x ?y () =
  object(self)
    inherit [full_item] container_item ~class_ ?name ?props ?group ?x ?y () as super
    (**/**)
    val mutable hbounds = Bounds.empty
    val mutable vbounds = Bounds.empty
    (**/**)

    method as_group = (self :> group)
    method items = List.map (fun c -> c.Container.data) children

    (** Removes all items from group. *)
    method clear =
      List.iter (fun i -> self#remove_item i) self#items

    (**/**)
    method! set_geometry g =
      Log.debug (fun m -> m "%s#set_geometry %a" self#me G.pp g);
      super#set_geometry g

    method! to_canvas_coords ~x ~y =
      let (x, y) = g.x + g_inner.x + x, g.y + g_inner.y + y in
      match group with
      | None -> (x,y)
      | Some g -> g#to_canvas_coords ~x ~y
    (**/**)

    method of_canvas_coords ~x ~y =
      let (x, y) = x - g.x - g_inner.x, y - g.y + g_inner.y in
      match group with
      | None -> (x,y)
      | Some g -> g#of_canvas_coords ~x ~y

    (**/**)
    method! freeze =
      super#freeze ;
      List.iter (fun c -> c.Container.data#freeze) children

    method! unfreeze =
      super#unfreeze ;
      List.iter (fun c -> c.Container.data#unfreeze) children

    method! get_item_at ~x:x_ ~y:y_ =
      if G.inside ~x:x_ ~y:y_ g then
        (
         let x = x_ - g.x - g_inner.x in
         let y = y_ - g.y - g_inner.y in
         let rec iter = function
         | [] -> Some (self#as_full_item)
         | c :: q ->
             match c.Container.data#get_item_at ~x ~y with
             | None -> iter q
             | x -> x
         in
         iter children
        )
      else
        None

    method! get_leaf_items_at acc ~x:x_ ~y:y_ =
      if G.inside ~x:x_ ~y:y_ g then
        (
         let x = x_ - g.x - g_inner.x in
         let y = y_ - g.y - g_inner.y in
         List.fold_left
           (fun acc c -> c.Container.data#get_leaf_items_at acc ~x ~y)
           acc children
        )
      else
        acc

    method! on_items_at f ~x:x_ ~y:y_ =
      Log.debug (fun m -> m "%s#on_items_at ~x:%d ~y:%d" self#me x_ y_);
      let x = x_ - g.x - g_inner.x in
      let y = y_ - g.y - g_inner.y in
      let rec iter = function
      | [] -> Lwt.return_false
      | c :: q ->
          let i = c.Container.data in
          Log.debug (fun m -> m "%s#on_items_at i=%s i#g=%a x=%d, y=%d"
            self#me i#me G.pp i#geometry x y);
          if G.inside ~x ~y i#geometry then
            match%lwt f i with
            | false -> i#on_items_at f ~x ~y
            | true -> Lwt.return_true
          else
            iter q
      in
      iter children
    (**/**)

    method width =
      let w = match Bounds.max_elt_opt hbounds with
        | None -> 0
        | Some (w,_) -> w
      in
      super#min_width_ + w
    method height =
      let h = match Bounds.max_elt_opt vbounds with
      | None -> 0
      | Some (h,_) -> h
      in
      super#min_height_ + h


    (**/**)
    method! private min_width_ = self#width
    method! private min_height_ = self#height

    method update_bounds ~recur =
      if recur then
        List.iter (fun c -> c.Container.data#update_bounds ~recur) children;
      self#update_hbound;
      self#update_vbound;
      self#set_geometry { g with w = self#width ; h = self#height }

    method update_hbound =
      hbounds <- List.fold_left
        (fun acc c ->
          let i = c.Container.data in
          let g = i#geometry in Bounds.add (g.G.x + g.w, i#id) acc)
        Bounds.empty children;

    method update_vbound =
      vbounds <- List.fold_left
        (fun acc c ->
          let i = c.Container.data in
          let g = i#geometry in Bounds.add (g.G.y + g.h, i#id) acc)
        Bounds.empty children

    method set_item_hbound (oldg:G.t) id (g:G.t) =
      (*if not frozen then*)
      (
       Log.debug (fun m -> m "%s#set_item_hbound oldg=%a id=%a g=%a"
          self#me G.pp oldg Oid.pp id G.pp g);
       let my_oldg = self#geometry in
       let oldw = self#width in
       let s = Bounds.remove (oldg.x + oldg.w, id) hbounds in
       hbounds <- Bounds.add (g.x + g.w, id) s;
       self#set_geometry { my_oldg with w = self#width };
       let w = self#width in
       Log.debug (fun m -> m "%s#set_item_hbound oldw=%d, w=%d, id=%a\nbounds=%s"
          self#me oldw w Oid.pp id
            (String.concat ", "
             (List.map
              (fun (x,id) -> Printf.sprintf "(%d,%s)" x (Oid.to_string id))
                (Bounds.elements hbounds))
            ));
       if oldw <> w then
         match group with
         | Some group ->
             group#set_item_hbound my_oldg self#id self#geometry;
             if not frozen then
               self#need_render ~layer:(self#get_p Props.layer) g
         | None -> if not frozen then super#need_resize
      )

    method set_item_vbound (oldg:G.t) id (g:G.t) =
      (*if not frozen then*)
      (
       Log.debug (fun m -> m "%s#set_item_vbound oldg=%a id=%a g=%a"
          self#me G.pp oldg Oid.pp id G.pp g);
       let my_oldg = self#geometry in
       let oldh = self#height in
       let s = Bounds.remove (oldg.y + oldg.h, id) vbounds in
       vbounds <- Bounds.add (g.y + g.h, id) s;
       self#set_geometry { my_oldg with h = self#height };
       let h = self#height in
       Log.debug (fun m -> m "%s#set_item_vbound oldh=%d, h=%d, id=%a\nbounds=%s"
          self#me oldh h Oid.pp id
            (String.concat ", "
             (List.map
              (fun (y,id) -> Printf.sprintf "(%d,%s)" y (Oid.to_string id))
                (Bounds.elements vbounds))
            ));
       if oldh <> h then
         match group with
         | Some group -> group#set_item_vbound my_oldg self#id self#geometry
         | None -> if not frozen then super#need_resize
      )
    (**/**)

    (** Adds the given item to group. The item's coordinates
      can be set with [x] and [y] optional arguments. *)
    method add_item ?x ?y (i:full_item) =
      match super#add i#coerce i with
      | false -> ()
      | true ->
         i#set_group (Some (self:>group));
          match x, y with
          | None, None -> i#set_xy (i#x, i#y)
          | Some x, None -> i#set_x x
          | None, Some y -> i#set_y y
          | Some x, Some y -> i#set_xy (x, y)

    (** Removes the given item from the group. *)
    method remove_item (i:full_item) =
      Log.debug (fun m -> m "%s#remove_item %s" self#me i#me);
      match super#remove i#coerce with
      | false -> ()
      | true ->
          let oldg = self#geometry in
          i#set_group_none ;
          let _g = i#geometry in
          self#update_bounds ~recur:false;
          let newg = self#geometry in
          Log.info (fun m -> m "%s#remove_item %s: oldg=%a, newg=%a" self#me i#me
            G.pp oldg G.pp newg);
          match group with
          | Some g when oldg <> newg ->
            g#set_item_vbound oldg self#id newg;
            g#set_item_hbound oldg self#id newg
            (*self#need_render ~layer:(self#get_p Props.layer) g*)
          | _ ->
            ()

    (**/**)
    val mutable child_need_render_cb = (None : (layer:Layer.t -> G.t -> unit) option)
    method set_child_need_render_cb x = child_need_render_cb <- x
    method child_need_render ~layer geom =
      if not frozen then
        (
         Log.debug (fun m -> m "%s#child_need_render: geom=%a"
            self#me G.pp geom);
         let g = self#geometry in
         let g_inner = self#g_inner in
         let geom = { geom with
             G.x = geom.G.x + g.x + g_inner.x;
             y = geom.y + g.y + g_inner.y;
           }
         in
         self#need_render ~layer geom
        )

    method need_render ~layer geom =
      match group with
      | Some group -> group#child_need_render ~layer geom
      | None ->
          match child_need_render_cb with
          | None -> ()
          | Some f -> f ~layer geom
    (**/**)

    (** Raises the given item above all the other items of the group. *)
    method raise (item : full_item) =
      match self#widget_data item#coerce with
      | None -> ()
      | Some data ->
          super#remove item#coerce ;
          ignore(super#add item#coerce data)

  end

(** An item to draw a rectangle.
   Default [class_] is ["canvas_rect"]. Width and height of the
   rectangle can be specified with optional arguments [w] and [h]. *)
class rect ?(class_="canvas_rect") ?name ?props ?group ?x ?y ~w ~h () =
  object(self)
    inherit full_item ~class_ ?name ?props ?group ?x ?y () as super
    (**/**)
    val mutable h = h
    val mutable w = w
    method private set_width w_ = w <- w_
    method private set_height h_ = h <- h_
    method! private min_width_ = w
    method! private min_height_ = h
    method! max_width = Some w
    method! max_height = Some h

    method! geometry = { g with h ; w }
    method! set_geometry g = super#set_geometry { g with h ; w }
    (**/**)

    method resize ?w ?h () =
      Option.iter self#set_width w;
      Option.iter self#set_height h;
      self#set_geometry g ;
      self#move ();(* to update parent's bounds *)

    (**/**)
    method! is_leaf_widget = true
    (**/**)
    initializer
      self#set_geometry { g with w ; h };
  end

(** A {!Text.class-label} as item. Default [class_] is ["canvas_label"]. *)
class label ?(class_="canvas_label") ?name ?props ?group ?x ?y () =
  object(self)
    inherit Text.label ~class_ ?name ?props () as widget
    inherit item ?group ?x ?y () as super
    method as_full_item = (self :> full_item)
  end

(** A {!Text.class-glyph} as item. Default [class_] is ["canvas_glyph"]. *)
class glyph ?(class_="canvas_glyph") ?name ?props ?group ?x ?y () =
  object(self)
    inherit Text.glyph ~class_ ?name ?props () as widget
    inherit item ?group ?x ?y () as super
    method as_full_item = (self :> full_item)
  end

(** A {!Pack.class-box} as item..*)
class ['a] box ?class_ ?name ?props ?group ?x ?y () =
  object(self)
    inherit ['a] Pack.box ?class_ ?name ?props () as widget
    inherit item ?group ?x ?y () as super
    method as_full_item = (self :> full_item)
    method set_geometry g = widget#set_geometry g
  end

(** A {!Pack.class-paned} as item. Default [class_] is ["canvas_paned"].*)
class paned ?class_ ?name ?props ?group ?x ?y () =
  object(self)
    inherit Pack.paned ?class_ ?name ?props () as widget
    inherit item ?group ?x ?y () as super
    method as_full_item = (self :> full_item)
    method set_geometry g = widget#set_geometry g
  end

(** A {!Bin.class-fixed_size} as item. *)
class fixed_size ?(class_="canvas_fixed_size") ?name ?props ?group ?x ?y ?w ?h () =
  object(self)
    inherit Bin.fixed_size ~class_ ?name ?props ?w ?h () as widget
    inherit item ?group ?x ?y () as super
    method as_full_item = (self :> full_item)
    method set_geometry g = widget#set_geometry g
  end

(** A {!Flex.class-flex} as item. Default [class_] is ["canvas_flex"]. *)
class ['a] flex ?(class_="canvas_flex") ?name ?props ?group ?x ?y () =
  object(self)
    inherit ['a] Flex.flex ~class_ ?name ?props () as flex
    inherit item ?group ?x ?y () as super
    method as_full_item = (self :> full_item)
    method set_geometry g = flex#set_geometry g
  end

(** {2:convenient_functions Convenient functions to create items}

These convenient functions are used to create items.
They all take [x] and [y] optional arguments to
specify the coordinates of the created item, and
[group] to specify the group the item belongs to.
Other arguments are {{!Widget.widget_arguments}generic widget arguments}
or arguments specific to the kind of item.
*)

(** Creates a {!class-group}. *)
let group ?class_ ?name ?props ?group ?x ?y () =
  new group ?class_ ?name ?props ?group ?x ?y ()

(** Creates a {!class-rect}. *)
let rect ?class_ ?name ?props ?group ?x ?y ~w ~h () =
  new rect ?class_ ?name ?props ?group ?x ?y ~w ~h ()

(** Creates a {!class-label}. *)
let label ?class_ ?name ?props ?group ?x ?y text =
  let t = new label ?class_ ?name ?props ?group ?x ?y () in
  t#set_text text ;
  t

(** Creates a {!class-glyph}. *)
let glyph ?class_ ?name ?props ?group ?x ?y gly =
  let t = new glyph ?class_ ?name ?props ?group ?x ?y () in
  t#set_glyph gly ;
  t

(** Creates a {!class-box} with vertical orientation. Default
  [class_] is ["canvas_hbox"]. *)
let vbox ?(class_="canvas_vbox") ?name ?props ?group ?x ?y () =
  let w = new box ~class_ ?name ?props ?group ?x ?y () in
  w#set_orientation Props.Vertical;
  w

(** Creates a {!class-box} with horizontal orientation. Default
  [class_] is ["canvas_hbox"]. *)
let hbox ?(class_="canvas_hbox") ?name ?props ?group ?x ?y () =
  let w = new box ~class_ ?name ?props ?group ?x ?y () in
  w#set_orientation Props.Horizontal ;
  w

(** Creates a {!class-paned} with vertical orientation. Default
  [class_] is ["canvas_vpaned"]. *)
let vpaned ?(class_="canvas_vpaned") ?name ?props ?group ?x ?y () =
  let w = new paned ~class_ ?name ?props ?group ?x ?y () in
  w#set_orientation Props.Vertical;
  w

(** Creates a {!class-paned} with horizontal orientation. Default
  [class_] is ["canvas_hpaned"]. *)
let hpaned ?(class_="canvas_hpaned") ?name ?props ?group ?x ?y () =
  let w = new paned ~class_ ?name ?props ?group ?x ?y () in
  w#set_orientation Props.Horizontal ;
  w

(** Creates a {!class-fixed_size}.*)
let fixed_size ?class_ ?name ?props ?group ?x ?y ?w ?h () =
  new fixed_size ?class_ ?name ?props ?group ?x ?y ?w ?h ()

(** Creates a {!class-flex}.*)
let flex ?class_ ?name ?props ?group ?x ?y
  ?(orientation=Props.Horizontal)
    ?justification ?items_alignment ?content_alignment
    ?inter_space ?wrap ?wrap_on_break ?collapse_spaces  () =
  let w = new flex ?class_ ?name ?props ?group ?x ?y () in
  w#set_orientation orientation ;
  Option.map w#set_justification justification ;
  Option.map w#set_items_alignment items_alignment ;
  Option.map w#set_content_alignment content_alignment ;
  Option.map w#set_inter_space inter_space ;
  Option.map w#set_wrap wrap ;
  Option.map w#set_wrap_on_break wrap_on_break ;
  Option.map w#set_collapse_spaces collapse_spaces;
  w

(** {2 Canvas widget} *)

class canvas ?(class_="canvas") ?name ?props () =
  object(self)
    inherit Widget.widget ~class_ ?name ?props () as super

    (**/**)
    val mutable root = new group ()
    (**/**)

    (** Returns root group. *)
    method root = root

    (** Sets root group. *)
    method set_root r =
      root#set_child_need_render_cb None ;
      root <- r ;
      r#set_parent ?with_rend:with_renderer (Some self#coerce) ;
      root#set_child_need_render_cb
        (Some self#child_need_render)

    (** Freezes, removes all items from root group and unfreezes. *)
    method clear =
      self#freeze ;
      root#clear ;
      self#unfreeze

    (** Returns item at the given coordinates, if any. Coordinates have
        the same origin as the canvas coordinates. *)
    method get_item_at ~x ~y =
      let g = root#geometry in
      root#get_item_at ~x:(x - g.x - g_inner.x) ~y:(y - g.y - g_inner.y)

    (** Get leaf items at the given coordinates. Coordinates have
        the same origin as the canvas coordinates. *)
    method get_leaf_items_at ~x ~y =
      let g = root#geometry in
      let items = root#get_leaf_items_at []
        ~x:(x - g.x - g_inner.x)
        ~y:(y - g.y - g_inner.y)
      in
      Log.debug (fun m -> m "%s#get_leaf_items => %s"
        self#me (String.concat ", " (List.map (fun i -> i#me) items)));
      items

    (**/**)
    method! wtree = Widget.N (self#coerce, [root#wtree])

    val mutable frozen_to_render = (None: G.t option)
    method! freeze =
      frozen_to_render <- Some G.zero;
      root#freeze
    method! unfreeze =
      match frozen_to_render with
      | None -> ()
      | Some rg ->
          frozen_to_render <- None;
          root#unfreeze;
          root#update_bounds ~recur:true;
          self#need_resize;
          self#need_render ~layer:(self#get_p Props.layer) rg

    method! child_need_render ~layer geom =
      let geom =
        let g = self#geometry in
        let g_inner = self#g_inner in
        { geom with
          x = geom.x + g.x + g_inner.x ;
          y = geom.y + g.y + g_inner.y }
      in
      match frozen_to_render with
      | None -> self#need_render ~layer geom
      | Some rg -> frozen_to_render <- Some (G.union rg geom)

    method! private min_width_ =
      let w = super#min_width_ + root#width in
      Log.debug (fun m -> m "%s#min_width => %d" self#me w);
      w
    method! private min_height_ =
      let h = super#min_height_ + root#height in
      Log.debug (fun m -> m "%s#min_height => %d" self#me h);
      h

    method! render_me ~layer rend ~offset:(x,y) rg =
      let (x,y) = (x + g.x + g_inner.x, y + g.y + g_inner.y) in
      let rg = G.translate ~x:(-g.x-g_inner.x) ~y:(-g.y-g_inner.y) rg in
      root#render ~layer rend ~offset:(x,y) rg

    method! set_geometry g =
      let g = { g with w = max self#min_width g.w ; h = max self#min_height g.h } in
      Log.debug (fun m -> m "%s#set_geometry with g=%a" self#me G.pp g);
      super#set_geometry g

    (* propagate only events with position = Some (mouse events)
       or keyboard related events (sent to the widget only if
       is_focus property is true. *)
    method! on_sdl_event_down ~oldpos pos e =
      if self#sensitive then
        let b =
          let pass_event =
            match Sdl.Event.(enum (get e typ)) with
            | `Key_down | `Key_up | `Text_input | `Text_editing -> true
            | _ -> pos <> None
          in
          if pass_event then
            let f (x,y) = (x - g.x - g_inner.x , y - g.y - g_inner.y) in
            let child_oldpos = Option.map f oldpos in
            let child_pos = Option.map f pos in
            root#on_sdl_event_down ~oldpos:child_oldpos child_pos e
          else
            false
        in
        Log.warn (fun m -> m "%s#on_sdl_event_down => return %b" self#me b);
        match b with
        | true -> true
        | false -> self#on_sdl_event pos e
      else
        false

    method! release_focus =
      match root#release_focus with
      | true ->
          self#set_p Props.is_focus false ;
          self#set_p Props.has_focus false ;
          true
      | _ -> false

    method! set_has_focus b =
      match super#set_has_focus b with
      | true -> true
      | false -> root#set_has_focus b

    method! grab_focus ?(last=false) () =
      Log.debug (fun m -> m "%s#grab_focus ~last:%b" self#me last);
      if self#visible then
        match self#get_p Props.focusable with
        | true ->
            Log.debug (fun m -> m "%s#grab_focus: i'm focusable" self#me);
            let b = match self#get_focus with
              | None -> false
              | Some _ -> true
            in
            Log.debug (fun m -> m "%s#grab_focus: get_focus => %b" self#me b);
            b
        | _ ->
            match self#get_p Props.can_focus with
            | false -> false
            | true -> root#grab_focus ~last ()
      else
        false

    initializer
      self#set_root root;
  end


(** Convenient function to create a {!class-canvas}.
   See {!Widget.widget_arguments} for arguments. *)
let canvas ?class_ ?name ?props ?pack () =
  let w = new canvas ?class_ ?name ?props () in
  Widget.may_pack ?pack w#coerce ;
  w

