(** Utils for managing keys. *)

open Tsdl

(** [key_is_mod k] returns whether [k] is a modifier. *)
val key_is_mod : Sdl.keymod -> bool

(** A keyboard state corresponds to a key state if the specified
     [key] is pressed and the active modifiers after applying [mask]
     are equal to [mods]. [mods] is not used directly but instead
     a predicate is built to handle modifiers having left and right
     variants. [mods] is kept only for printing or storing purpose.
     [mask] is used to keep only modifiers of interest. For example
     not containing [Sdl.Kmod.caps] results in ignoring the state
     of Caps Lock when testing. Modifiers with left and right
     variants are not concerned by the mask since they are
     handled in the predicate. *)
type keystate = {
    key : Sdl.keycode;
    mask : Sdl.keymod;
    mods : Sdl.keymod;
    pred : Sdl.keymod -> bool;
}

val compare_keystate : keystate -> keystate -> int
val keystate_equal : keystate -> keystate -> bool
val pp_keystate : Format.formatter -> keystate -> unit

(** Set default mask used by {!val-keystate} if no
  mask is specified in argument. Default is {!Tsdl.Sdl.Kmod.none}.
  [Sdl.Kmod.(reserved+mode)]. *)
val set_default_keymask : Sdl.keymod -> unit
val default_keymask : unit -> Sdl.keymod

(** Set default modifiers used by {!val-keystate} if no
  modifier is specified in argument. Default is
  [Sdl.Kmod.(reserved+mode)]. *)
val set_default_keymods : Sdl.keymod -> unit
val default_keymods : unit -> Sdl.keymod

(** Create a {!type-keystate} from the given key.
  Optional parameters are:
  {ul
  {- [mask] to set the modifiers whose state will be ignored. If not
     specified, {!default_keymask} is used.}
  {- [mods] to set the required modifiers. If not
     specified, {!default_keymods} is used.}
  }
*)
val keystate :
  ?mask:Sdl.keymod ->
  ?mods:Sdl.keymod ->
  Sdl.keycode ->
  keystate

(** [match_keys keystate ~key ~kmod] returns [true] if the given
  [key] and modifier [kmod] maths the given [keystate]. *)
val match_keys : keystate -> key:Sdl.keycode -> kmod:Sdl.keymod-> bool

val keystate_of_string : string -> keystate
val string_of_mods : Sdl.keymod -> string
val string_of_keystate : keystate -> string

val keystate_ocf_wrapper : keystate Ocf.Wrapper.t
val keystate_list_ocf_wrapper : keystate list Ocf.Wrapper.t

val keystate_list_of_string : string -> keystate list
val string_of_keystate_list : keystate list -> string

val is_mod_pressed : Sdl.keymod -> bool
val shift_pressed : unit -> bool
val ctrl_pressed : unit -> bool
