(** Objects.

  Class {!class-o} is the base class for widget and other classes which
  can trigger events and hold properties {!Props.props}.
*)

(** Introduce the [Prop_changed] property to register callbacks called
  when the value of a given property changed. Previous and current
  values of the property are passed to the callback. *)
type _ Events.ev +=
| Prop_changed : 'a Props.prop -> (prev: 'a option -> now: 'a -> unit) Events.ev


(** Object id. Each object of {!class-o} is given a unique id. *)
type id = Oid.t

(** An object. See {!Widget.widget_arguments} for
  [class_], [name] and [props] arguments.

  Registered callbacks are unregistered when object is destroyed.
*)
class o class_ ?name ?props () =
  let props =
    let p = Theme.of_class ?name class_ in
    match props with
    | None -> p
    | Some p2 -> Props.merge p p2
  in
  let name = match name with None -> class_ | Some s -> s in
  let id = Oid.gen () in
  object(self)
    (** The unique id of the object. *)
    method id = id

    (** Coercion to {!class-o}. *)
    method as_o = (self :> o)

    (** Returns a string representation of the object's name and id.*)
    method me = Printf.sprintf "[%s/%s]" name (Oid.to_string self#id)

    (**/**)
    val mutable props = Props.dup props
    (**/**)

    (** Returns the object properties. This is not a copy and should
      not be modified directly except if you know what you're doing.
      To set a property value, use the [set_*] methods provided by
      inheriting classes. If no such method is available for a
      property, you can use the [set_p] method below.*)
    method props = props

    (** [o#get_p p] returns value of property [p] in properties of [o].
      If no value for [p] is set in [o] and [p] has no default value,
      then {!Misc.type-error.Missing_prop} is raised.
    *)
    method get_p : 'a. 'a Props.prop -> 'a = fun p -> Props.get props p

    (** [o#opt_p p] returns value of property [p] in properties of [o],
      if such a value is set. *)
    method opt_p : 'a. 'a Props.prop -> 'a option = fun p -> Props.opt props p

    (** [o#set_p p v] sets value [v] to property [p] in properties of [o].
      Optional argument [propagate] (defaults: [false]) can be used
      to indicat that this change must be propagated to children objects.
      Class {!class-o} has no children, but inheriting classes may have and
      will override this method to implement propagation.
      The [Prop_changed] event is triggered if the value of [p] changed
      (properties are defined with a comparison function, allowing
       to detect a change).
    *)
    method set_p : 'a. ?propagate:bool -> 'a Props.prop -> 'a -> unit =
      fun ?(propagate=false) p v ->
        match Props.update props p v with
        | None -> (* no change *) ()
        | Some prev ->
            try
              Log.debug
                (fun m -> m "%s property %S updated => %a (propagate=%b)"
                   self#me (Props.name p) (Props.pp_prop p) v
                     propagate);
              self#on_prop_changed p ~prev v ;

            with Not_found ->
                Log.err (fun  m -> m "Not found! %s" (Printexc.get_backtrace()))

    (**/**)
    method on_prop_changed : 'a. 'a Props.prop -> prev:'a option -> 'a -> unit =
      fun p ~prev now ->
        let cbs = Events.get self#id (Prop_changed p) in
        List.iter (fun cb -> cb ~prev ~now) cbs
    (**/**)

    (** [o#set_props ?propagate props] calls [o#set_p ?propagate] for
      each pair (property, value) in [props]. *)
    method set_props ?(propagate=false) props =
      Props.fold
        (fun p v () -> self#set_p ~propagate p v)
        props
        ()

    (**/**)
    val mutable callback_ids = ([] : Events.callback_id list)
    (**/**)

    (** [o#connect event cb] registers [cb] to be called when [event]
      is triggered by [o]. Returns a callback id which can be used
      to unregister the callback.*)
    method connect : 'a. 'a Events.ev -> 'a -> Events.callback_id =
      fun e cb ->
        let id = Events.register self#id e cb in
        callback_ids <- id :: callback_ids;
        id

    (** [disconnect id] unregisters the callback with the given [id]. *)
    method disconnect = Events.unregister

    (**/**)
    method trigger_event : 'a. ('a -> bool) Events.ev -> 'a -> bool =
      fun ev v ->
        Log.debug (fun m -> m "%s#trigger_event %s" self#me
          (Printexc.to_string (Obj.magic ev)));
        match Events.get self#id ev with
        | [] -> false
        | cbs ->
          List.fold_left
            (fun acc cb -> let b = cb v in (b || acc))
              false cbs

    method trigger_event_unit : 'a. ('a -> bool) Events.ev -> 'a -> unit =
      fun ev v -> ignore(self#trigger_event ev v)

    method trigger_unit_event : 'a. ('a -> unit) Events.ev -> 'a -> unit =
      fun ev v ->
        match Events.get self#id ev with
        | [] -> ()
        | cbs -> List.iter
            (fun f ->
               try f v
               with e -> Log.err (fun m -> m "%s" (Printexc.to_string e)))
              cbs

    method destroy = List.iter Events.unregister callback_ids
    initializer
      Gc.finalise (fun o -> o#destroy) self

  end

