(** Boxes to pack widgets. *)

open Misc
open Tsdl
open Widget
open Container

[@@@landmark "auto"]

(** Property ["inter_padding"] use to specify a space between packed children.
  Default is [0]. *)
let inter_padding = Props.int_prop ~inherits:false ~default:0 "inter_padding"

(**/**)
let set : 'a. Widget.widget -> 'a Props.prop -> 'a option -> unit = fun w -> fun p -> function
    | None -> ()
    | Some v -> w#set_p p v
(**/**)

(** {2 Stacking boxes} *)

(** A {!class-box} is a {!Container.container} which stacks its children,
  giving each one more or less space according to its size and properties.
  Widgets can be stacked horizontally or vertically (depending on
  the {!Props.val-orientation} property).*)
class ['a] box ?(class_="box") ?name ?props () =
  object(self)
    inherit ['a option] Container.container ~class_ ?name ?props () as super

    (** {2 Properties} *)

    method orientation = self#get_p Props.orientation
    method set_orientation = self#set_p Props.orientation

    method inter_padding = self#get_p inter_padding
    method set_inter_padding = self#set_p inter_padding

    (** {2 Children } *)

    (** [b#widget_data w] returns {!Container.child.data} data associated to the given
      child widget [w], if [w] is a child of [b] and it has data associated. *)
    method widget_data = super#widget_data

    (** [b#widget_index w] returns 0-based index of [w] in [b]'s children,
      if [w] is a child of [b]. *)
    method widget_index = super#widget_index

    (** [o#children_widgets] returns the list of children widget of [o]. *)
    method children_widgets = List.map (fun c -> c.widget) self#children

    (** [o#reorder_child w pos] moves child widget [w] to new position [pos]
       (if possible). *)
    method reorder_child w pos = super#reorder_child w pos

    (**/**)
    method! private min_width_ =
      super#min_width_ +
        match self#orientation with
        | Props.Horizontal ->
            let (w, nb, mright) =
              List.fold_left
                (fun (acc_w, acc_nb, prev_margin) c ->
                 if c.widget#visible && c.widget#hexpand >= 0
                   then
                     let (mleft,mright) = c.widget#hmargin in
                     let w = c.widget#min_width - mright - mleft + (max prev_margin mleft) in
                     (acc_w + w, acc_nb + 1, mright)
                   else (acc_w, acc_nb, prev_margin))
                (0, 0, 0) children
            in
            w + mright + ((max 0 (nb - 1)) * self#inter_padding)
        | Vertical ->
            List.fold_left
              (fun acc c ->
                 if c.widget#visible && c.widget#hexpand >= 0
                 then max acc c.widget#min_width
                 else acc)
              0 children

    method! private min_height_ =
      super#min_height_ +
        match self#orientation with
        | Horizontal ->
            List.fold_left
              (fun acc c ->
                 if c.widget#visible && c.widget#vexpand >= 0
                 then max acc c.widget#min_height
                 else acc)
              0 children
        | Vertical ->
            let (h, nb, mbottom) =
              List.fold_left
                (fun (acc_h, acc_nb, prev_margin) c ->
                   if c.widget#visible && c.widget#vexpand >= 0
                   then
                    let (mtop,mbottom) = c.widget#vmargin in
                     let h = c.widget#min_height - mbottom - mtop + (max prev_margin mtop) in
                     (acc_h + h, acc_nb + 1, mbottom)
                   else (acc_h, acc_nb, prev_margin))
              (0, 0, 0) children
            in
            h + mbottom + ((max 0 (nb - 1)) * self#inter_padding)

    method! max_width =
      match self#hexpand with
      | n when n <> 0 -> None
      | _ ->
          let f op (acc_exists_none, max_w) c =
            if c.widget#visible then
              match c.widget#max_width with
              | None -> (true, max_w)
              | Some w -> (acc_exists_none, op max_w w)
            else
              (acc_exists_none, max_w)
          in
          match
            match self#orientation with
            | Horizontal ->
                List.fold_left (f (+)) (false, 0) children
            | Vertical ->
                List.fold_left (f max) (false, 0) children
          with
          | true, _  -> None
          | false, max_w -> Some (max_w + super#min_width_)

    method! max_height =
      match self#vexpand with
      | n when n <> 0 -> None
      | _ ->
          let f op (acc_exists_none, max_h) c =
            if c.widget#visible then
              match c.widget#max_height with
              | None -> (true, max_h)
              | Some h -> (acc_exists_none, op max_h h)
            else
              (acc_exists_none, max_h)
          in
          match
            match self#orientation with
            | Horizontal ->
                List.fold_left (f max) (false, 0) children
            | Vertical ->
                List.fold_left (f (+)) (false, 0) children
          with
          | true, _  -> None
          | false, max_h -> Some (max_h + super#min_height_)

    (**/**)

    (** [o#pack w] adds widget [w] to [o]. Optional parameters are:
      {ul
       {- [pos] indicates a position to insert [w]; default is to
          append [w] to children.}
       {- [hexpand] (resp. [vexpand]) sets {!Props.hexpand} (resp.
          {!Props.vexpand}) property of [w] to the given value. }
       {- [hfill] (resp. [vfill]) sets {!Props.hfill} (resp.
          {!Props.vfill}) property of [w] to the given value. }
       {- [data] associates the given value to [w]. All data associated
          to children must have the same type (this type is the type
          parameter of class {!class-box}. }
      }
       To allocate space for children widgets, the following algorithm
       is applied (here for horizontal packing; for vertical packing,
       replace width by height and [hexpand] by [vexpand]):
       {ol
       {- The minimum widths of children are summed. This sum is deducted
          from the width allocated to [o].}

       {- The remaining space is then distributed among children,
          according to their [hexpand] value.
          [0] means that widget does not require more than its
             minimal width.
          A positive value [p] means that the widget requires
          [p] shares of the remaining space. The remaining space
          is divided by the total number of shares and each widget
          is given as width its minimal width + the width corresponding
          to its required shares.
          For example, if three widgets of same minimum size have [hexpand]
          values of [3], [2] and [1], the first widget will have half
          of the available width, the second will have one third and
           the last on sixth.
       }
       {- When allocated width of each widget is computed, each widget
          is given this width (if value of [hfill] is [true]) of just its
          minimum width (if it is [false]).}
      }
    *)
    method pack ?pos ?hexpand ?vexpand ?hfill ?vfill ?data w =
      Log.debug (fun m -> m "%s#add %s" self#me w#me);
      set w Props.hexpand hexpand ;
      set w Props.vexpand vexpand ;
      set w Props.hfill hfill ;
      set w Props.vfill vfill ;
      match super#add ?pos w data with
      | false -> ()
      | true -> if w#visible then self#need_resize

    (** [o#unpack w] removes child widget [w] from [o]. *)
    method unpack (w : Widget.widget) =
      match super#remove w with
      | false -> ()
      | true -> if w#visible then self#need_resize

    (** [o#unpack_all ~destroy] removes all children from [o]. [destroy]
       indicates whether to call [#destroy] on children after removing. *)
    method unpack_all ~destroy =
      match self#children_widgets with
      | [] -> ()
      | l ->
          let old_nr = ignore_need_resize in
          self#ignore_need_resize ;
          List.iter
            (fun w ->
              self#unpack w;
              if destroy then w#destroy
            )
            l;
          if not old_nr then
            (self#handle_need_resize ;
             self#need_resize)

    (**/**)
    method! set_geometry geom =
      super#set_geometry geom ;
      Log.debug (fun m -> m "%s#set_geometry g=%a g_inner=%a"
         self#me G.pp g G.pp g_inner);
      Log.debug (fun m -> m "%a" Widget.pp_widget_tree self#wtree);
      (match self#orientation with
       | Horizontal -> self#set_geometry_horizontal geom
       | Vertical -> self#set_geometry_vertical geom
      );
      self#need_render ~layer:(self#get_p Props.layer) g

    method private set_geometry_horizontal geom =
      let w = self#min_width in
      let w = max w geom.w in
      let w = w - super#min_width_ in
      let visib_children = self#visible_children in
      let ip = self#inter_padding in
      Log.debug (fun m -> m "%s#set_geometry inter_padding=%d" self#me ip);
      let w_avail =
        let spaces =
          let (w, mright) =
            List.fold_left (fun (acc_w, prev) w ->
               let (mleft,mright) = w#hmargin in
               (acc_w + (max (max ip mleft) prev), mright))
              (0, 0) visib_children
          in
          w + mright
        in
        max 0 (w - spaces)
      in
      let m = Packer.compute w_avail
        (fun w ->
           let (mleft, mright) = w#hmargin in
           w#min_width - mleft - mright)
          (fun w ->
             match w#max_width with
             | None -> None
             | Some mw ->
                 let (mleft, mright) = w#hmargin in
                 Some (mw - mleft - mright)
          )
          (fun w -> w#hexpand)
          visib_children
      in
      let _ = List.fold_left
        (fun (x,prev_margin) (wid:Widget.widget) ->
           let t = Oid.Map.find wid#id m in
           let h = match wid#max_height with
             | None -> g_inner.h
             | Some h ->
                 Log.debug (fun m ->
                    m "%s#set_geometry wid=%s wid#max_height=%d" self#me wid#me h);
                 min h g_inner.h
           in
           Log.debug (fun m -> m "%s#set_geometry for %s h=%d" self#me wid#me h);
           let m = wid#margin in
           let h =
             let minh = wid#min_height in
             if wid#vfill then
               max minh h
             else
               minh
           in
           let h = h - m.top - m.bottom in
           Log.debug (fun m -> m "%s#set_geometry for %s h=%d vfill=%b"
              self#me wid#me h wid#vfill);
           let w =
             if wid#hfill then
               t.current
             else
               t.min
           in
           let x =
             let mleft = if x = 0 then m.left else max ip m.left in
             x + max mleft prev_margin
           in
           let geo = {
               G.x; y = m.top;
               w ; h }
           in
           wid#set_geometry geo ;
           geo.x + t.current, m.right
        )
          (0,0)
          visib_children
      in
      ()

    method private set_geometry_vertical geom =
      let h = self#min_height in
      let h = max h geom.h in
      let h = h - super#min_height_ in
      let visib_children = self#visible_children in
      let ip = self#inter_padding in
      let h_avail =
        let spaces =
          let (h, mbottom) =
            List.fold_left (fun (acc_h, prev) w ->
               let (mtop,mbottom) = w#vmargin in
               (acc_h + (max (max ip mtop) prev), mbottom))
              (0, 0) visib_children
          in
          h + mbottom
        in
        max 0 (h - spaces)
      in
      let m = Packer.compute h_avail
        (fun w ->
           let (mtop,mbottom) = w#vmargin in
           w#min_height - mtop - mbottom)
          (fun w ->
             match w#max_height with
             | None -> None
             | Some h ->
                 let (mtop,mbottom) = w#vmargin in
                 Some (h - mtop - mbottom))
          (fun w -> w#vexpand)
          visib_children
      in
      let _ = List.fold_left
        (fun (y, prev_margin) (wid:Widget.widget) ->
           let t =
             match Oid.Map.find_opt wid#id m with
             | None -> Log.err (fun pr -> pr
                    "%s#set_geometry: widget with id=%s not found in widget list:[%s]"
                      self#me (Oid.to_string wid#id)
                      (String.concat "; "
                       (List.map
                        (fun (id,_) -> Printf.sprintf "%s"
                           (Oid.to_string id))
                          (Oid.Map.bindings m)
                       )
                      )
                 );
                 raise Not_found
             | Some x -> x
           in
           let m = wid#margin in
           let w = match wid#max_width with
             | None -> g_inner.w
             | Some w -> min w g_inner.w
           in
           let w =
             let minw = wid#min_width in
             if wid#hfill then
               max minw w
             else
               minw
           in
           let w = w - m.left - m.right in
           let h =
             if wid#vfill then
               t.current
             else
               t.min
           in
           let y =
             let mtop = if y = 0 then m.top else max ip m.top in
             y + max mtop prev_margin
           in
           let geo = {
               G.x = m.left ; y;
               w; h }
           in
           wid#set_geometry geo ;
           geo.y + t.current, m.bottom
        )
          (0, 0) visib_children
      in
      ()
  end


(** Convenient function to create a {!class-box}.
  The [class_] argument defaults to ["vbox"] if [orientation]
  is [Vertical] or ["hbox"] if [orientation] is [Horizontal].
  See {!Widget.widget_arguments} for other arguments. *)
let box ~orientation ?class_ ?name ?props ?pack () =
  let class_ = match class_ with
    | None -> Printf.sprintf "%cbox"
        (match orientation with Props.Vertical -> 'v' | Horizontal -> 'h')
    | Some s -> s
  in
  let w = new box ~class_ ?name ?props () in
  w#set_orientation orientation ;
  Widget.may_pack ?pack w ;
  w

(** Same as {!val-box} but orientation is already fixed to [Horizontal].*)
let hbox ?class_ ?name ?props ?pack () =
  box ~orientation:Horizontal ?class_ ?name ?props ?pack ()

(** Same as {!val-box} but orientation is already fixed to [Vertical].*)
let vbox ?class_ ?name ?props ?pack () =
  box ~orientation:Vertical ?class_ ?name ?props ?pack ()

(** {2 Paned widgets} *)


(** A handle position, defined either as percentage or absolute position. *)
type handle_position = [`Percent of float | `Absolute of int]

let handle_position_wrapper =
  let to_json ?with_doc = function
  | `Percent v -> `Float v
  | `Absolute v -> `Int v
  in
  let from_json ?def = function
  | `Float v -> `Percent (max 0. (min 100. v))
  | `Int v -> `Absolute (max 0 v)
  | json -> Ocf.invalid_value json
  in
  Ocf.Wrapper.make to_json from_json

(**/**)
module THandle_positions = struct
    type t = handle_position option list
    let compare = Stdlib.compare
    let wrapper = Some Ocf.Wrapper.(list (option handle_position_wrapper))
  end
module PHandle_positions = Props.Add_prop_type(THandle_positions)
(**/**)

(** Property ["handle_positions"], to store the positions of the handles
  in a {!class-paned} widget. *)
let handle_positions : handle_position option list Props.prop = PHandle_positions.mk_prop
  ~after:[Props.Resize]
  ~default:[]
  ~inherits:false
  "handle_positions"

(** How to define handle position when the user moves it: as percentage
  or absolute value. This changes the way handle positions are updated
  when the {!class-paned} widget is resized: with [`Percent], ratios between
  children will be kept, but with [`Absolute] the handles will remain at
  the same position. *)
type user_handle_positionning = [`Percent | `Absolute]

let user_handle_positionning_wrapper : user_handle_positionning Ocf.wrapper =
  let to_json ?with_doc = function
  | `Percent -> `String "percent"
  | `Absolute -> `String "absolute"
  in
  let from_json ?def = function
  | (`String s) as json ->
      (match String.lowercase_ascii s with
      | "percent" -> `Percent
      | "absolute" -> `Absolute
      | _ -> Ocf.invalid_value json)
  | json -> Ocf.invalid_value json
  in
  Ocf.Wrapper.make to_json from_json

(**/**)
module TUser_handle_positionning = struct
    type t = user_handle_positionning
    let compare = Stdlib.compare
    let wrapper = Some user_handle_positionning_wrapper
  end
module PUser_handle_positionning = Props.Add_prop_type(TUser_handle_positionning)
(**/**)

(** Property ["user_handle_positionning"]. *)
let user_handle_positionning : user_handle_positionning Props.prop =
  PUser_handle_positionning.mk_prop
    ~default:`Percent ~inherits:false "user_handle_positionning"

(**/**)
let default_handle_props =
  let p = Props.empty () in
  Props.(
    set p width 2;
    set p fg_color Color.grey ;
  );
  p
(**/**)

(** Property ["paned_handle_props"] to define appearance of
  handles in {!class-paned} widget.
  Default set {!Props.val-width} to [2] and {!Props.val-fg_color} to
  {!Color.grey}.
*)
let handle_props = Props.props_prop ~after:[Props.Resize]
  ~default:default_handle_props ~inherits:false "paned_handle_props"

(** Property ["paned_user_set_handle_live_update"] defines whether
  to update paned widget on each move of a handle by user ([true]),
  or wait for the user to end moving the handle ([false]).
  Default is [true]. It may be useful to set it to [false] when computation
  of children sizes may take some time. *)
let user_set_handle_live_update =
  Props.bool_prop ~after:[Props.Resize]
    ~default:true ~inherits:false "paned_user_set_handle_live_update"
(*
let honor_child_min_size = Props.bool_prop ~after:[Props.Resize]
  ~default:true ~inherits:false "paned_honor_child_min_size"
*)

(** Paned widget.

  Contains children widget separated by handles. Horizontal (or vertical
  if orientation is [Vertical]) space allocated to each child depends
  on handle positions. The first handle position defines how to distribute
  paned widget space between first and second children. The second handle
  position indicate how to distribube the remaining space between the
  second and third childreb, and so on.*)
class paned ?(class_="paned") ?name ?props () =
  object(self)
    inherit [unit] Container.container ~class_ ?name ?props () as super

    (**/**)
    (* coordinates of handles relative to g_inner *)
    val mutable g_handles = ([] : G.t list)
    val mutable handle_cursor =
      let> c = Sdl.(create_system_cursor System_cursor.size_ns) in
      c
    val mutable cursor_on_handle = false
    val mutable state_machine = Misc.empty_state_machine
    (**/**)

    (** {2 Properties} *)

    method orientation = self#get_p Props.orientation
    method set_orientation = self#set_p Props.orientation

    method handle_props = self#get_p handle_props
    method set_handle_props = self#set_p handle_props

    method handle_positions = self#get_p handle_positions
    method set_handle_positions = self#set_p handle_positions

    method user_handle_positionning = self#get_p user_handle_positionning
    method set_user_handle_positionning = self#set_p user_handle_positionning

    method user_set_handle_live_update = self#get_p user_set_handle_live_update
    method set_user_set_handle_live_update = self#set_p user_set_handle_live_update
(*
    method honor_child_min_size = self#get_p honor_child_min_size
    method set_honor_child_min_size = self#set_p honor_child_min_size
*)

    (** {2 Children} *)

    (** [o#children_widgets] returns the list of children widget of [o]. *)
    method children_widgets = List.map (fun c -> c.widget) self#children


    (** [o#reorder_child w pos] moves child widget [w] to new position [pos]
       (if possible). *)
    method reorder_child w pos = super#reorder_child w pos

    (**/**)
    method! private min_width_ = super#min_width_ +
      self#min_handles_width +
        (let vchildren = self#visible_children in
         let op = match self#orientation with
         | Props.Horizontal -> (+)
         | Props.Vertical -> max
         in
         List.fold_left (fun acc w -> op acc w#min_width) 0 vchildren
        )

    method! private min_height_ = super#min_height_ +
      self#min_handles_height +
        (let vchildren = self#visible_children in
         let op = match self#orientation with
           | Props.Horizontal -> max
           | Props.Vertical -> (+)
         in
         List.fold_left (fun acc w -> op acc w#min_height) 0 vchildren
        )

    method private min_handles_width =
      let vchildren = self#visible_children in
      match self#orientation with
      | Horizontal ->
        let w = Props.(get self#handle_props width) in
        (max 0 (List.length vchildren - 1)) * w
      | Vertical -> 0

    method private min_handles_height =
      let vchildren = self#visible_children in
      match self#orientation with
      | Horizontal ->
          let w = Props.(get self#handle_props width) in
          (max 0 (List.length vchildren - 1)) * w
      | Vertical -> 0

    method! max_width =
      match self#hexpand with
      | n when n <> 0 -> None
      | _ ->
          let f op (acc_exists_none, max_w) c =
            if c.widget#visible then
              match c.widget#max_width with
              | None -> (true, max_w)
              | Some w -> (acc_exists_none, op max_w w)
            else
              (acc_exists_none, max_w)
          in
          match
            match self#orientation with
            | Horizontal ->
                List.fold_left (f (+)) (false, 0) children
            | Vertical ->
                List.fold_left (f max) (false, 0) children
          with
          | true, _  -> None
          | false, max_w -> Some (max_w + super#min_width_ + self#min_handles_width)

    method! max_height =
      match self#vexpand with
      | n when n <> 0 -> None
      | _ ->
          let f op (acc_exists_none, max_h) c =
            if c.widget#visible then
              match c.widget#max_height with
              | None -> (true, max_h)
              | Some h -> (acc_exists_none, op max_h h)
            else
              (acc_exists_none, max_h)
          in
          match
            match self#orientation with
            | Horizontal ->
                List.fold_left (f max) (false, 0) children
            | Vertical ->
                List.fold_left (f (+)) (false, 0) children
          with
          | true, _  -> None
          | false, max_h -> Some (max_h + super#min_height_ + self#min_handles_height)

    method private set_geometry_horizontal =
      let handle_w = Props.(get self#handle_props width) in
      let rec iter acc vchildren handle_pos x =
        match vchildren with
        | [] -> List.rev acc
        | [w] ->
            w#set_geometry { G.x ; y = 0 ; w = g_inner.w - x ; h = g_inner.h };
            List.rev acc
        | w1 :: qw ->
            let remain_w = max 0 (g_inner.w - x - handle_w) in
            let (target_w, qpos) =
              match handle_pos with
              | [] as q
              | None :: q -> remain_w / 2, q
              | (Some (`Absolute p)) :: q -> min remain_w p, q
              | (Some (`Percent p)) :: q ->
                  truncate (float remain_w *. (p /. 100.)), q
            in
            let g1 =
              let w =
                let w = max w1#min_width target_w in
                match w1#max_width with
                | None -> w
                | Some maxw -> min w maxw
              in
              { G.x ; y = 0 ; w ; h = g_inner.h }
            in
            w1#set_geometry g1;
            let gh = { G.x = g1.x + g1.w ; y = 0 ; w = handle_w ; h = g_inner.h } in
            iter (gh :: acc) qw qpos (gh.x + gh.w)
      in
      g_handles <- iter [] self#visible_children self#handle_positions 0

    method private set_geometry_vertical =
      let handle_h = Props.(get self#handle_props width) in
      let rec iter acc vchildren handle_pos y =
        match vchildren with
        | [] -> List.rev acc
        | [w] ->
            w#set_geometry { G.x = 0 ; y ; w = g_inner.w ; h = g_inner.h - y };
            List.rev acc
        | w1 :: qw ->
            let remain_h = max 0 (g_inner.h - y - handle_h) in
            let (target_h, qpos) =
              match handle_pos with
              | [] as q
              | None :: q -> remain_h / 2, q
              | (Some (`Absolute p)) :: q -> min remain_h p, q
              | (Some (`Percent p)) :: q ->
                  truncate (float remain_h *. (p /. 100.)), q
            in
            let g1 =
              let h =
                let h = max w1#min_height target_h in
                match w1#max_height with
                | None -> h
                | Some maxh -> min h maxh
              in
              { G.x = 0 ; y ; w = g_inner.w ; h }
            in
            w1#set_geometry g1;
            let gh = { G.x = 0; y = g1.y + g1.h ; w = g_inner.w ; h = handle_h } in
            iter (gh :: acc) qw qpos (gh.y + gh.h)
      in
      g_handles <- iter [] self#visible_children self#handle_positions 0

    method! set_geometry geom =
      super#set_geometry geom ;
      Log.debug (fun m -> m "%s#set_geometry g=%a" self#me G.pp g);
      (match self#orientation with
      | Horizontal -> self#set_geometry_horizontal
      | Vertical -> self#set_geometry_vertical);
      self#need_render ~layer:(self#get_p Props.layer) g

    (**/**)

    (** [o#pack w] adds widget [w] to [o]. Optional parameter
        [pos] indicates a position to insert [w]; default is to
          append [w] to children. *)
    method pack ?pos w =
      Log.debug (fun m -> m "%s#add %s" self#me w#me);
      match super#add ?pos w () with
      | false -> ()
      | true -> if w#visible then self#need_resize

    (** [o#unpack w] removes child widget [w] from [o]. *)
    method unpack (w : Widget.widget) =
      match super#remove w with
      | false -> ()
      | true -> if w#visible then self#need_resize

    (** [o#unpack_all ~destroy] removes all children from [o]. [destroy]
       indicates whether to call [#destroy] on children after removing. *)
    method unpack_all ~destroy =
      match self#children_widgets with
      | [] -> ()
      | l ->
          let old_nr = ignore_need_resize in
          self#ignore_need_resize ;
          List.iter
            (fun w ->
              self#unpack w;
              if destroy then w#destroy
            )
            l;
          if not old_nr then
            (self#handle_need_resize ;
             self#need_resize)

    (**/**)

    (* coordinates relative to g_inner *)
    method private handle_from_coords ~x ~y =
      let rec iter n = function
      | [] -> None
      | gh :: q ->
          let gh = match self#orientation with
            | Horizontal -> G.enlarge ~w:2 gh
            | Vertical -> G.enlarge ~h:2 gh
          in
          if G.inside ~x ~y gh then Some n else iter (n+1) q
      in
      iter 0 g_handles

    method private user_set_handle_pos n ~x ~y =
      let rec iter acc prev_bound i (lgh : G.t list) lpos =
        match lgh with
        | [] -> List.rev acc
        | gh :: qgh when i = n ->
            let bound = match self#orientation with
              | Horizontal ->
                  max 0 ((min x g_inner.w) - prev_bound)
              | Vertical ->
                  max 0 ((min y g_inner.h) - prev_bound)
            in
            let bound =
              match self#user_handle_positionning with
              | `Absolute -> `Absolute bound
              | `Percent ->
                  let limit =
                    match self#orientation with
                    | Horizontal -> g_inner.w
                    | Vertical -> g_inner.h
                  in
                  `Percent ((float bound /. float (limit - prev_bound)) *. 100.)
            in
            let lpos =
              match lpos with
              | [] -> [Some bound]
              | _ :: lpos -> (Some bound) :: lpos
            in
            (List.rev acc @ lpos)
        | gh :: qgh ->
            let prev_bound =
              match self#orientation with
              | Horizontal -> gh.x + gh.w + 1
              | Vertical -> gh.y + gh.h + 1
            in
            let (acc, lpos) =
              match lpos with
              | [] -> None :: acc, []
              | p :: q -> (p :: acc), q
            in
            iter acc prev_bound (i+1) qgh lpos
      in
      let new_pos = iter [] 0 0 g_handles self#handle_positions in
      Log.debug (fun m -> m "%s#user_set_handle_pos n=%d x=%d y=%d handle_positions=%a"
        self#me n x y (Props.pp_prop handle_positions) new_pos);
      self#set_handle_positions new_pos

    method! on_sdl_event_down ~oldpos pos ev =
      if self#sensitive then
        match state_machine.f pos ev with
        | false -> super#on_sdl_event_down ~oldpos pos ev
        | true -> true
      else
        false

    method on_mouse_leave =
      (match state_machine.state () with
       | `Moving_handle _ -> state_machine.set_state `Base
       | _ -> ()
      );
      super#on_mouse_leave

    method private restore_cursor = Sdl.set_cursor self#top_widget#cursor

    method state_on_event state pos ev =
      match state, pos, Sdl.Event.(enum (get ev typ)) with
      | `Base, Some(x,y), `Mouse_motion ->
          (
           (* change cursor if needed *)
           (*Log.warn (fun m -> m "%s#state_on_event mouse_motion g=%a x=%d y=%d"
              self#me G.pp g x y);*)
           let (x, y) = self#to_g_inner_coords ~x ~y in
           match self#handle_from_coords ~x ~y, cursor_on_handle with
           | None, false -> None
           | None, true ->
               self#restore_cursor ;
               cursor_on_handle <- false;
               None
           | Some _, true -> None
           | Some _, false ->
               cursor_on_handle <- true ;
               Sdl.set_cursor (Some handle_cursor);
               None
          )
      | `Moving_handle n, Some (x, y), `Mouse_motion ->
          if G.inside ~x ~y g then
            (
             let (x, y) = self#to_g_inner_coords ~x ~y in
             if self#user_set_handle_live_update then
               self#user_set_handle_pos n ~x ~y;
             Some (`Moving_handle n, true)
            )
          else
            (
             cursor_on_handle <- false ;
             self#restore_cursor ;
             Some (`Base, false)
            )
      | `Base, Some (x,y), `Mouse_button_down ->
          let button = Sdl.Event.(get ev mouse_button_button) in
          if button = 1 then
            let (x, y) = self#to_g_inner_coords ~x ~y in
            match self#handle_from_coords ~x ~y with
            | None -> None
            | Some n -> Some (`Moving_handle n, true)
          else
            None
      | `Moving_handle n, Some (x, y), `Mouse_button_up ->
          let (x, y) = self#to_g_inner_coords ~x ~y in
          self#user_set_handle_pos n ~x ~y;
          Some (`Base, false)
      | (`Base|`Moving_handle _), _, _ -> None

    method! render_me ~(layer:Layer.t) (rend:Sdl.renderer) ~offset:(x,y) (rg:G.t) =
      let off_x = g.x + g_inner.x in
      super#render_me ~layer rend ~offset:(x,y) rg ;
      let off_y = g.y + g_inner.y in
      let offset = (x + off_x, y + off_y) in
      let rg = G.translate ~x:(-off_x) ~y:(-off_y) rg in
      List.iter (self#render_handle ~layer rend ~offset rg) g_handles

    method render_handle ~layer rend ~offset:(x,y) rg gh =
      match G.inter rg gh with
      | None -> ()
      | Some clip ->
          let clip = G.translate ~x ~y clip in
          let f rend =
            let gh = G.translate ~x ~y gh in
            let col = Props.(get self#handle_props fg_color) in
            Render.fill_rect rend (Some gh) col
          in
          Render.with_clip rend (G.to_rect clip) f

    method! destroy =
      super#destroy ;
      Sdl.free_cursor handle_cursor

    initializer
      state_machine <- Misc.mk_state_machine `Base self#state_on_event ;
      ignore(self#connect (Object.Prop_changed Props.orientation)
        (fun ~prev ~now ->
           Sdl.free_cursor handle_cursor ;
           let> c = Sdl.(create_system_cursor
            System_cursor.(
             match self#orientation with
             | Horizontal -> size_we
             | Vertical -> size_ns))
          in
          handle_cursor <- c
       ))

  end

let paned orientation ?class_ ?name ?user_set_handle_live_update ?props ?pack () =
  let class_ = match class_ with
    | None -> Printf.sprintf "%cpaned"
        (match orientation with Props.Vertical -> 'v' | Horizontal -> 'h')
    | Some s -> s
  in
  let w = new paned ~class_ ?name ?props () in
  w#set_orientation orientation ;
  Option.iter w#set_user_set_handle_live_update user_set_handle_live_update ;
  Widget.may_pack ?pack w ;
  w

let hpaned = paned Horizontal
let vpaned = paned Vertical
