(** Tags associated to characters in {!Textbuffer} and {!Textview}. *)

(** {2 Text tags} *)

(** Tags. *)
module T : sig

  type t (** A tag id *)

  (** [create name] creates a new tag with given [name].
    If a tag with the same name already exists, a warning
    is issued and the id of the previous tag is returned. *)
  val create : string -> t

  val compare : t -> t -> int
  val equal : t -> t -> bool

  val name : t -> string

  val pp : Format.formatter -> t -> unit

  (** [get_or_create name] returns tag with given [name] if
    if exists, or else create it. *)
  val get_or_create : string -> t
  end

module TMap: Map.S with type key = T.t
module TSet: Misc.Idset with type id = T.t

val pp_list : Format.formatter -> T.t list -> unit
val pp_set : Format.formatter -> TSet.t -> unit

(** {2 Logging tags}

See {!Textlog}.
*)

val tag_debug : T.t
val tag_info : T.t
val tag_warning : T.t
val tag_error : T.t
val tag_app : T.t

(** {2 Syntax highlighting tags} *)

(** Tags associated to {!Higlo.Lang.token} and used for syntax
  highlighting in textviews. *)
module Lang : sig
    val bcomment : T.t
    val constant : T.t
    val directive : T.t
    val escape : T.t
    val id : T.t
    val keyword0 : T.t
    val keyword1 : T.t
    val keyword2 : T.t
    val keyword3 : T.t
    val keyword4 : T.t
    val lcomment : T.t
    val numeric : T.t
    val string : T.t
    val symbol0 : T.t
    val symbol1 : T.t
    val title0 : T.t
    val title1 : T.t
    val title2 : T.t
    val title3 : T.t
    val title4 : T.t
    val title5 : T.t
    val title6 : T.t
    val tag_of_token : Higlo.Lang.token -> T.t option * int
  end

(** {2 Themes} *)

(** Such a theme associates properties to text tags. These properties
  are used by {!Textview.class-textview} widget to display characters according
  to their tags. A theme also has properties, to get default values for
  properties not associated to text tags.
  *)
module Theme : sig
    type t (** A theme. *)

    val compare : t -> t -> int
    val equal : t -> t -> bool
    val pp : Format.formatter -> t -> unit

    (** Initialization. It is called by {!App.val-init}. *)
    val init : unit -> unit

    (** The default tag theme, with name ["default"].*)
    val default :t

    (** Get theme name. *)
    val name : t -> string

    (** Get theme by name, if it exists. *)
    val get_by_name : string -> t option

    (** [create name] creates a new theme.
       Optional arguments:
       {ul
        {- [tags]: initialize theme with a mapping of text tags to properties,}
        {- [props]: global properties of theme.}
       }
       If a theme with the same [name] already exists, a warning is issued
      and the existing theme is returned.
   *)
    val create : ?tags:Props.t TMap.t -> ?props:Props.t -> string -> t

    (** Get properties of theme. *)
    val props : t -> Props.t

    (** Set properties of theme. *)
    val set_props : t -> Props.t -> unit

    (** [set_tag theme tag props] set [tag] properties in [theme]. *)
    val set_tag : t -> T.t -> Props.t -> unit

    (** [opt_props theme tag] returns properties of [tag] in [theme], if tag is
       defined in theme. *)
    val opt_props : t -> T.t -> Props.t option

    (** [tag_props theme tag] gets the properties of [tag] in [theme].
      If [tag] has no properties, then an empty {!Props.t} is associated
      to this tag in [theme] and is returned. *)
    val tag_props : t -> T.t -> Props.t

    (** [set_tag_prop theme tag p v] sets value [v] to property [p] of [tag]
      in [theme]. *)
    val set_tag_prop : t -> T.t -> 'a Props.prop -> 'a -> unit

    (** [merge_tag_props theme set props] merge tags properties with initial
      [props] properties. Properties of first created tags are merged first.
      This function is used by {!Textview.class-textview} widget to compute
      the properties of a character according to its tags. *)
    val merge_tag_props : t -> TSet.t -> Props.t -> Props.t

    (** [get_or_create name] gets theme with given [name] if it exists,
      or else create it.
      Optional arguments:
      {ul
       {- [tags]: a mapping of text tags to properties. It is set as initial
          map if theme is created. If theme already exists, properties of
          each tag in the theme are merged with the properties of the same
          tag in the map.}
       {- [props]: global properties of theme. If the theme already exists,
          its properties are merged with [props].}
      }
    *)
    val get_or_create : ?tags:Props.t TMap.t ->
      ?props:Props.t -> string -> t

    (** Property ["tagtheme"] to store the name of the tag theme. Default is
       ["default"]. Inherited.*)
    val prop : string Props.prop

    (** [to_json t] returns a JSON representation of [t]. *)
    val to_json : t -> Yojson.Safe.t

    (** [from_json name json] creates theme [name] and sets properties
       and text tags properties from [json] description. The function
       uses {!get_or_create} so an existing theme with the same name
       will be modified by this function.*)
    val from_json : string -> Yojson.Safe.t -> t
  end